# -*- encoding: utf-8 -*-

from config_manager import safe_get_host
from command_execute import report, kill_transform, cancel_transform, set_job_threads
from threading import Thread
import time
import ztq_core

class CommandThread(Thread):
    """ 监视命令队列，取得命令, 执行命令"""

    def __init__(self, worker_name=''):
        super(CommandThread, self).__init__()
        self.login_time = int(time.time())
        self.worker_name = worker_name or safe_get_host('server', 'alias')

    def init(self):
        """ 开机初始化工作 """
        reboot = False
        worker_state = ztq_core.get_worker_state()
        if self.worker_name in worker_state:
            # 重启，读取服务器配置信息
            reboot = True
        # 记录系统日志
        system_log = ztq_core.get_system_log_queue()
        system_log.push(dict( host=safe_get_host('server', 'alias'),
                              alias=self.worker_name,
                              type=reboot and 'reboot' or 'power',
                              timestamp=self.login_time,))
        # 报告机器状态
        worker_state[self.worker_name] = report(self.login_time)

    def run(self):
        self.init()
        # 监听指令
        commands = ztq_core.get_command_queue(self.worker_name)
        while True:
            try:
                command = commands.pop()
                if command['command'] == 'report':
                    worker_state = ztq_core.get_worker_state()
                    worker_state[self.worker_name] = report(self.login_time)
                elif command['command'] == 'updatedriver':
                    # TODO
                    #async_drive_config()
                    pass
                elif command['command'] == 'updateworker':
                    queue = ztq_core.get_worker_config()
                    set_job_threads(queue[self.worker_name])
                elif command['command'] == 'kill':
                    kill_transform(pid=command['pid'], timestamp=command['timestamp'])
                elif command['command'] == 'cancel':
                    cancel_transform(pid=command['pid'], timestamp=command['timestamp'])
            except KeyboardInterrupt:
                import os
                # 实际上调用的是command_execute.clear_thread
                os.sys.exitfunc()
                os._exit(0)



# -*- encoding: utf-8 -*-
import threading
import time

import ztq_core

class BufferThread(threading.Thread):

    def __init__(self, config):
        """ cofnig: {'job-0':{'thread_limit': 50},,,}
        """
        super(BufferThread, self).__init__()
        self.config = config
        self._stop = False

    def run(self):

        if not self.config: return 

        while not self._stop:
            for buffer_name, buffer_config in self.config.items():

                # 需要停止
                if self._stop: return

                self.buffer_queue = ztq_core.get_buffer_queue(buffer_name)
                self.task_queue = ztq_core.get_task_queue(buffer_name)
                self.buffer_name = buffer_name
                self.task_queue_limit = int(buffer_config['thread_limit'])

                while True:
                    try:
                        self.start_job()
                        break
                    except ztq_core.ConnectionError:
                        time.sleep(3)

            time.sleep(1)

    def start_job(self):
        over_task_limit = self.task_queue_limit - len(self.task_queue)

        # 这个任务可能还没有push上去，服务器就挂了，需要在重新push一次
        if getattr(self, 'buffer_task', None):
            self.push_buffer_task()

        # 相关的任务线程，处于繁忙状态
        if over_task_limit <= 0:
            return

        # 不繁忙，就填充满
        else:
            while over_task_limit > 1:

                # 需要停止
                if self._stop: return

                # 得到一个任务
                self.buffer_task = self.buffer_queue.pop(timeout=-1)
                if self.buffer_task is None:
                    return 

                self.push_buffer_task()

                self.buffer_task = None
                over_task_limit -= 1

    def push_buffer_task(self):
        if 'runtime' not in self.buffer_task:
            self.buffer_task['runtime'] = {}

        ztq_core.push_runtime_task(self.buffer_name, self.buffer_task)

    def stop(self):
        self._stop = True



# -*- encoding: utf-8 -*-
import threading
import time, sys
import traceback
import logging

from config_manager import safe_get_host, CONFIG
import ztq_core

thread_context = threading.local()
logger = logging.getLogger("ztq_worker")
QUEUE_TIMEOUT = 30

def report_job(pid=-1, comment='', **kw):
    """ 报告当前转换进程信息 """
    if not hasattr(thread_context, 'job'):
        return  # 如果不在线程中，不用报告了

    job = thread_context.job

    # 报告转换状态
    job['process'].update({'pid': pid,
                        'start':int(time.time()),
                        'comment':comment})
    if kw:
        job['process'].update(kw)

    # 写入状态
    job_state = ztq_core.get_job_state(job['runtime']['worker'])
    job_state[job['runtime']['thread']] = job

def report_progress(**kw):
    """ 报告当前转换进程信息 """
    if not hasattr(thread_context, 'job'):
        return  # 如果不在线程中，不用报告了

    job = thread_context.job

    if not 'progress_callback' in job: return
    # 报告转换进度
    progress_func  = ztq_core.task_registry[job['pcallback']]
    progress_args = job.get('pcallback_args', [])
    progress_kw  = job.get('pcallback_kw', {})
    progress_kw.update(kw)
    progress_func(*progress_args, **progress_kw)

class JobThread(threading.Thread):
    """ 监视一个原子队列，调用转换引擎取转换
        转换结果记录转换队列，转换出错需要记录出错日志与错误队列
    """
    def __init__(self, queue_name, sleep_time, from_right=True):
        super(JobThread, self).__init__()
        self.queue_name = queue_name
        self.sleep_time = sleep_time
        self.from_right = from_right # 读取服务器队列的方向，从左边还是右边
        # _stop 为 True 就会停止这个线程
        self._stop = False
        self.start_job_time = 0  #  记录任务开始时间

    def run(self):
        """ 阻塞方式找到任务，并自动调用"""
        # 如果上次有任务在运行还没结束，重新执行
        jobs = ztq_core.get_job_state(safe_get_host('server', 'alias'))
        if self.name in jobs:
            self.start_job(jobs[self.name])

        # 队列批处理模式
        # batch_size: 批处理的阀值，达到这个阀值，就执行一次batch_func
        # batch_func: 
        #    1, 执行一批batch_size 大小的任务后，后续自动执行这个方法方法
        #    2, 执行一批小于batch_size 大小的任务后，再得不到任务，后续自动执行这个方法
        batch_config = CONFIG.get("batch_queue", {}).get(self.queue_name, {})
        batch_size = batch_config.get('batch_size', None) or -1
        batch_func = batch_config.get('batch_func', None) or (lambda *args, **kw: -1)

        run_job_index = 0
        queue_tiemout = QUEUE_TIMEOUT
        # 循环执行任务
        while not self._stop:
            try:
                task = ztq_core.pop_task(
                        self.queue_name, 
                        timeout=queue_tiemout, 
                        from_right=self.from_right
                        )
            except ztq_core.ConnectionError:
                print 'ERROR: Not connected the server\n'
                task = None
                time.sleep(3)

            if task is None: 
                # 没有后续任务了。执行batch_func
                if run_job_index > 0:
                    run_job_index = 0
                    queue_tiemout = QUEUE_TIMEOUT
                    batch_func()
                continue

            self.start_job(task)

            if batch_size > 0: 
                if run_job_index >= batch_size - 1:
                    # 完成了一批任务。执行batch_func
                    run_job_index = 0
                    queue_tiemout = QUEUE_TIMEOUT
                    batch_func()
                else:
                    run_job_index += 1
                    queue_tiemout = -1

            if self.sleep_time:
                time.sleep(self.sleep_time)

    def start_job(self, task):
        self.start_job_time = int(time.time())
        task['runtime'].update({'worker': safe_get_host('server', 'alias'),
                                'thread': self.getName(),
                                'start': self.start_job_time, })
        # 记录当前在做什么
        task['process'] = {'ident':self.ident}
        thread_context.job = task
        try:
            self.run_task = ztq_core.task_registry[task['func']]
            # started report
            report_job(comment='start the job')

            try:
                self.run_task(*task['args'], **task['kw'])
            except TypeError:
                # keyword must string is a bug in python
                if sys.version[:5] < '2.6.5':
                    raise Exception,("We not supported %s version of python,"
                        "Please update it to 2.6.5 or later."%sys.version[:5])
                else:
                    raise

            if task.get('callback', None):
                callback_args = task.get('callback_args', ())
                callback_kw = task.get('callback_kw', {})
                ztq_core.push_task(task['callback'], *callback_args, **callback_kw)

        except Exception, e:
            reason = traceback.format_exception(*sys.exc_info())
            # 将错误信息记录到服务器
            return_code = str(e.args[0]) if len(e.args) > 1 else 300
            task['runtime']['return'] = return_code
            task['runtime']['reason'] = reason[-11:]
            task['runtime']['end'] = int( time.time() )
            ztq_core.push_runtime_error(self.queue_name, task)
            # 错误回调
            if task.get('fcallback', None):
                callback_args = task.get('fcallback_args', ())
                callback_kw = task.get('fcallback_kw', {})
                callback_kw['return_code'] = return_code
                callback_kw['return_msg'] = reason[-1]
                ztq_core.push_task(task['fcallback'], *callback_args, **callback_kw)
            # 在终端打印错误信息
            #reason.insert(0, str(datetime.datetime.today()) + '\n')
            logger.error(''.join(reason))

        # 任务结束，记录日志
        task['runtime']['end'] = int( time.time() )
        if not 'reason' in task['runtime']:
            task['runtime']['return'] = 0
            task['runtime']['reason'] = 'success'
        ztq_core.get_work_log_queue().push(task)
        # 删除服务器的转换进程状态信息
        job_state = ztq_core.get_job_state(task['runtime']['worker'])
        del job_state[task['runtime']['thread']]
        self.start_job_time = 0

    def stop(self):
        """ 结束这个进程，会等待当前转换完成
            请通过JobThreadManager 来完成工作线程的退出，不要直接使用这个方法
        """
        self._stop = True




# -*- encoding: utf-8 -*-

import sys, os
from command_thread import CommandThread
from config_manager import init_config, get_configs, safe_get_host
from command_execute import init_job_threads, set_job_threads

import ztq_core

def main():
    """ 主函数 """
    conf_file = ''
    # 用户指定一个配置文件
    if len(sys.argv) > 1:
        conf_file = sys.argv[1]

    init_config(conf_file)
    server = get_configs('server')

    alias = safe_get_host('server', 'alias')
    active_config = server.get('active_config', 'false')

    # 动态注册task
    for module in server['modules'].split():
        try:
            __import__(module)
        except ImportError:
            modules = module.split('.')
            __import__(modules[0], globals(), locals(), modules[1])

    # 连结服务器
    ztq_core.setup_redis('default', host=server['host'], port=int(server['port']), db=int(server['db']))

    # 开启一个命令线程
    command_thread = CommandThread(worker_name=alias)

    sys.stdout.write('Starting server in PID %s\n'%os.getpid())

    worker_state = ztq_core.get_worker_state()
    if active_config.lower() == 'true' and command_thread.worker_name in worker_state:
        # 如果服务器有这个机器的配置信息，需要自动启动工作线程
        queue = ztq_core.get_worker_config()
        if command_thread.worker_name in queue:
            set_job_threads(queue[command_thread.worker_name])
    elif get_configs('queues'):
        # 把worker监视队列的情况上报到服务器
        queue_config = ztq_core.get_queue_config()
        # 如果配置有queues，自动启动线程监视
        job_threads = {}
        for queue_name, sleeps in get_configs('queues').items():
            job_threads[queue_name] = [
                    {'interval': int(sleep)} for sleep in sleeps.split(',')
                    ]
            if not queue_config.get(queue_name, []):
                queue_config[queue_name] = {'name':queue_name, 'title':queue_name, 'widget': 5}

        init_job_threads(job_threads)

    loggers = get_configs('log')
    initlog(
        loggers.get('key', 'ztq_worker'), 
        loggers.get('handler_file'), 
        loggers.get('level', 'ERROR'),
        )

    # 不是以线程启动
    command_thread.run()

def initlog(key, handler_file, level):
    import logging
    level = logging.getLevelName(level)
    format = '%(asctime)s %(message)s'
    if not handler_file:
        logging.basicConfig(level=level, format=format)
    else:
        logging.basicConfig(
                filename=handler_file, 
                filemode='a', 
                level=level, 
                format=format
                )



# -*- encoding: utf-8 -*-

#from zopen.transform import set_drive_config
from config_manager import safe_get_host, set_config
from job_thread_manager import JobThreadManager
from buffer_thread import BufferThread
from system_info import get_cpu_style, get_cpu_usage, get_mem_usage
import os
import sys
import traceback
import time
import ztq_core

# 管理工作线程, 添加线程、删除线程、保存信息
job_thread_manager = JobThreadManager()

# buffer 线程
buffer_thread_instance = None

def set_job_threads(config_dict):
    """ 根据配置信息和job_thread_manager.threads 的数量差来退出/增加线程
        剩下的修改queue_name, interval
    """
    tmp_jobs = job_thread_manager.threads.copy()
    config = []
    # 将config_dict转换格式为dicter = [{'queue':'q01', 'interval':6, }, ]
    for queue_name, values in config_dict.items():
        for value in values:
            dicter = dict( queue=queue_name )
            dicter.update(value)
            config.append(dicter)

    diff_job = len(config) - len(tmp_jobs)
    if diff_job > 0: # 需要增加线程
        for i in xrange(diff_job):
            conf = config.pop()
            job_thread_manager.add(conf['queue'], conf['interval'], conf.get('from_right', True))

    elif diff_job < 0: # 需要退出线程
        for key in tmp_jobs.keys():
            tmp_jobs.pop(key)
            job_thread_manager.stop(key)
            diff_job += 1
            if diff_job >= 0:
                break

    # 剩下的修改queue、interval、from_right, 如果有
    for index, job_thread in enumerate(tmp_jobs.values()):
        conf = config[index]
        job_thread.queue_name = conf['queue']
        job_thread.sleep_time = conf['interval']
        job_thread.from_right = conf.get('from_right', True)

def init_job_threads(config_dict):
    # 启动工作线程
    set_job_threads(config_dict)
    # 保存信息
    config = {}
    for queue_name, values in config_dict.items():
        sleeps = [str(x['interval']) for x in values]
        config[queue_name] = ','.join(sleeps)
    set_config(config, 'queues')

    # 将一些信息补全，让监视界面认为这个worker已经启动
    alias = safe_get_host('server', 'alias')
    # 初始化
    dispatcher_config = ztq_core.get_dispatcher_config()
    if not dispatcher_config: 
        dispatcher_config = {'queue_weight':{},'worker_weight':{}}
    # set dispatcher config
    worker_weight = dispatcher_config['worker_weight']
    if not worker_weight.get(alias, 0):
        worker_weight[alias] = 5 
        ztq_core.set_dispatcher_config(dispatcher_config)
    # set worker config
    worker_config = ztq_core.get_worker_config()
    worker_config[alias] = config_dict 

def set_dirve( from_mime, to_mime, conf):
    """ 根据驱动配置, 更改驱动参数 """
    #set_drive_config(from_mime, to_mime, conf)
    pass

def report(start_time):
    """ 转换器向服务器报告状态 """
    cpu_style = get_cpu_style()
    cpu_percent = get_cpu_usage() 
    mem_percent, mem_total = get_mem_usage()
    ip = safe_get_host('server', 'alias')

    traceback_dict = {}
    for thread_id, frame in sys._current_frames().items():
        traceback_dict[thread_id] = traceback.format_stack(frame)

    # 向服务器报告
    return dict( ip=ip,
            cpu_style=cpu_style, 
            cpu_percent=cpu_percent, 
            mem_total=mem_total, 
            mem_percent=mem_percent, 
            started=start_time, 
            timestamp=int(time.time()),
            traceback=traceback_dict,
            )

def kill_transform(pid, timestamp):
    """ 中止 转换 """
    kill(pid)

def cancel_transform(pid, timestamp):
    """ 取消 转换 """
    kill(pid)

if os.sys.platform != 'win':
    def kill(pid):
        """ kill process by pid for linux """
        # XXX 无法杀孙子进程
        kill_command = "kill -9 `ps --no-heading --ppid %s|awk '{print $1}'` %s" % (pid, pid)
        os.system(kill_command)
else:
    def kill(pid):
        """ kill process by pid for windows """
        kill_command = "taskkill /F /T /pid %s" % pid
        os.system(kill_command)

def start_buffer_thread(buffer_thread_config):
    """ 开启一个buffer队列线程，监视所有的buffer队列，
        根据buffer队列对应的job队列拥塞情况, 将buffer队列的任务合适的推送到相应的job队列
    """
    if not buffer_thread_config: return

    global buffer_thread_instance
    if buffer_thread_instance is not None:
        buffer_thread_instance.stop()

    buffer_thread = BufferThread(buffer_thread_config)
    buffer_thread.setDaemon(True)
    buffer_thread.start()

    buffer_thread_instance = buffer_thread
    sys.stdout.write('start a buffer thread. \n')

def clear_transform_thread(threads=None):
    """ clear job_threads and buffer_thread """
    threads = threads or job_thread_manager.threads
    names = threads.keys()
    job_threads = threads.values()

    # 退出buffer 线程
    if buffer_thread_instance is not None:
        buffer_thread_instance.stop()
        sys.stdout.write('wait the buffer thread stop...\n')

    # 将进程的stop 标志 设置为True
    map(job_thread_manager.stop, names)

    # 如果这个线程没有工作，只是在阻塞等待任务，就发送一个空的任务
    # 让这个线程立刻结束
    for job_thread in job_threads:
        if job_thread.start_job_time == 0:
            queue_name = job_thread.queue_name
            queue = ztq_core.get_task_queue(queue_name)
            queue.push('')

    # 等待线程退出
    for job_thread in job_threads:
        sys.stdout.write('wait the %s stop...\n'%job_thread.getName())
        job_thread.join(30)

import atexit
atexit.register(clear_transform_thread) # 系统退出后清理工作线程




# -*- encoding: utf-8 -*-

import os 

from ConfigParser import ConfigParser
from system_info import get_ip

# 读取配置文件（app.ini），保存到CONFIG中，实际使用的都是CONFIG
CONFIG = {}

def init_config(location=None):
    """ 初始化配置管理
    """
    cfg = ConfigParser()
    if location:
        cfg.read(location)
    else:
        local_dir = os.path.dirname(os.path.realpath(__file__))
        cfg.read( os.path.join(local_dir, 'config.cfg') )

    global CONFIG
    for section in cfg.sections():
        CONFIG[section] = {}
        for option in cfg.options(section):
            CONFIG[section][option] = cfg.get(section, option)

def set_config(config, section=''):
    """ 设置配置信息，默认设置全部，可选只设置某个节点 """
    global CONFIG
    if section:
        config = {section:config}

    for key, value in config.items():
        section = CONFIG.setdefault(key, {})
        section.update(value)

def get_configs(section):
    """ 得到节点全部的相关信息 """
    if CONFIG is None:
        return ''
    return CONFIG.get(section, {})

def get_config(section, option):
    """ 得到某个节点的某个值 """
    if CONFIG is None:
        return ''
    return CONFIG.get(section, {}).get(option, '')

def safe_get_host(section, option):
    """ 得到worker的的host name
        先找配置文件，是否配置了server节点下的alias 
        没有就返回worker机器的IP

        get_ip 方法已经缓存了结果
    """
    host = get_config(section, option)
    if not host:
        return get_ip()
    return host

def register_batch_queue(queue_name, batch_size, batch_func=None):
    """ 注册队列是批处理模式
        queue_name: 指定哪个队列为批处理模式
        batch_size: 整形
        batch_func: 方法对象

        可以让队列在完成batch_size 后，执行一次 batch_func
    """
    set_config(
            config={queue_name:{'batch_size':batch_size, 'batch_func':batch_func}}, 
            section="batch_queue",
            )



import os
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.txt')).read()
CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()

setup (
    name='ztq_worker',
    version='1.0',
    author = "xutaozhe",
    author_email = "xutaozhe@zopen.cn",
    description=u"Zopen Task Queue Worker",
    long_description=README + '\n\n' +  CHANGES,
    license = "MIT",
    keywords='Everydo queue async ztq_worker',
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        ],
    packages = ['ztq_worker'],
    package_dir={'ztq_worker': 'ztq_worker'},
    package_data={'ztq_worker': ['system_info/*.vbs'] },
    data_files=[('config', ['worker.ini']),],
    include_package_data = True,
    install_requires = [
        "ztq_core",
        ],
    entry_points = """\
      [console_scripts]
      ztq_worker = ztq_worker.main:main
      """,
    zip_safe = False,
)




from command_thread import CommandThread
from job_thread import report_progress, report_job
from config_manager import register_batch_queue
from command_execute import start_buffer_thread, init_job_threads




# -*- coding: utf-8 -*-

from job_thread import JobThread
import sys

class JobThreadManager:
    """ 管理工作线程， 开启/停止工作线程 """
    # 保存工作线程的信息
    # threads = {'thread-1':<object JobThread>, 'thread-2':<object JobThread>}
    threads = {}

    def add(self, queue_name, sleep_time, from_right=True):
        """ 开启一个工作线程 """
        job_thread = JobThread(queue_name, sleep_time, from_right)
        job_thread.setDaemon(True)
        job_thread.start()
        self.threads[job_thread.getName()] = job_thread
        sys.stdout.write(
                'start a job thread, name: %s,'
                ' ident: %s,'
                ' queue_name: %s\n'
                % (job_thread.getName(), job_thread.ident, queue_name)
                )

    def stop(self, job_name):
        """ 安全的停止一个工作线程
            正在转换中的时候，会等待转换完成后自动退出
        """
        if not job_name in self.threads:
            return
        #sys.stdout.write('stop %s job thread\n'% job_name)
        job_thread = self.threads[job_name]
        job_thread.stop()
        del self.threads[job_name]
